# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: utils.py 
@time: 2020/06/16
"""

# python packages
from functools import reduce

# 3rd-party packages
from loguru import logger


# self-defined packages
@logger.catch(reraise=True)
def compose(*funcs):
    """Compose arbitrarily many functions, evaluated left to right.
    Reference: https://mathieularose.com/function-composition-in-python/
    """
    # return lambda x: reduce(lambda v, f: f(v), funcs, x)
    if funcs:
        return reduce(lambda f, g: lambda *a, **kw: g(f(*a, **kw)), funcs)
    else:
        raise ValueError('Composition of empty sequence not supported.')



# -*- coding:utf-8 _*-
"""
@author: Maureen Hsu
@file: gpu_setting.py
@time: 2019/10/21
"""

# python packages

# 3rd-party packages
import tensorflow.keras as keras
from loguru import logger
import matplotlib.pyplot as plt


class LogHistory(keras.callbacks.Callback):
    def on_train_begin(self, logs=None):
        pass

    def on_epoch_end(self, epoch, logs=None):
        s = f"[Epoch {epoch+1}/{self.params['epochs']}]"
        for key, value in logs.items():
            s += f" {key}:{value:.4f}"

        logger.debug(s)


def plot_history(history, save_path=None):
    plt.plot(history.history['loss'])
    plt.plot(history.history['val_loss'])
    plt.title('model loss')
    plt.ylabel('loss')
    plt.xlabel('epoch')
    plt.legend(['train', 'val'], loc='upper left')
    if save_path:
        plt.savefig(save_path)

    plt.show()




def get_balanced_weight(dic_class):
    total = 0
    dic_weight = {}

    dic_class_idx = {klass: i for i, klass in enumerate(sorted(dic_class.keys()))}

    for klass in dic_class:
        fraction = 1. / dic_class[klass]
        dic_weight[dic_class_idx[klass]] = fraction
        total += fraction

    multiplier = 1.0 / total

    for klass in dic_weight:
        dic_weight[klass] = dic_weight[klass] * multiplier

    logger.debug("Class weights are {}".format(
        ", ".join("{}: {:.4f}".format(klass, weight) for klass, weight in dic_weight.items())))

    return dic_weight


def convert_tf1_model_to_tf2(tf2_model, tf1_weights, model_path=None, weight_path=None):
    tf2_model.load_weight(tf1_weights)
    if model_path is not None:
        tf2_model.save(model_path)
    if weight_path is not None:
        tf2_model.save_weights(weight_path)



import datetime
import multiprocessing as mp
import threading
import queue
import time

import cv2
from loguru import logger

now = datetime.datetime.now


class VideoYielderError(Exception):
    def __init__(self, msg):
        self.msg = msg


class VideoCaptureDaemon(threading.Thread):
    def __init__(self, URL, result_queue):
        super().__init__()
        self.daemon = True
        self.URL = URL
        self.result_queue = result_queue

    def run(self):
        self.result_queue.put(cv2.VideoCapture(self.URL))


def get_video_capture(URL, timeout=5):
    res_queue = queue.Queue()
    VideoCaptureDaemon(URL, res_queue).start()
    try:
        return res_queue.get(block=True, timeout=timeout)
    except queue.Empty:
        logger.error(
            'cv2.VideoCapture: could not grab input ({}). Timeout occurred after {:.2f}s'.format(URL, timeout))
        return None


class VideoCaptureNoQueue:
    def __init__(self, URL, timeout=5):
        self.frame = []
        self.status = False
        self.is_stop = False
        self.count = 0

        # 攝影機連接。
        cap = None
        while cap is None:
            cap = get_video_capture(URL, timeout=timeout)
            timeout += 1
        self.capture = cap

        self.start()

    def start(self):
        # 把程式放進子執行緒，daemon=True 表示該執行緒會隨著主執行緒關閉而關閉。
        threading.Thread(target=self.query_frame, daemon=True, args=()).start()

    def release(self):
        # 記得要設計停止無限迴圈的開關。
        self.is_stop = True

    def get(self, opt):
        return self.capture.get(opt)

    def read(self):
        # 當有需要影像時，再回傳最新的影像。
        return self.count, self.frame

    def query_frame(self):
        while not self.is_stop and self.capture.isOpened():
            self.status, self.frame = self.capture.read()
            self.count += 1

        self.frame = None  # if capture is stopped, put empty frame to send signal to mother process
        self.capture.release()


@logger.catch(reraise=True)
class VideoYielder:
    def __init__(self, path, n_frame_per_sec=None, timeout=5, auto_restart=True):
        self.path = path
        self.n_frame_per_sec = n_frame_per_sec
        self.timeout = timeout
        self.count = 0
        self.video_retry = 0
        self.cap = None
        self.fps = 0
        self.interval = 1
        self.last_time = None
        self.auto_restart = auto_restart

    def start(self):
        self.cap = VideoCaptureNoQueue(self.path, timeout=self.timeout)
        self.fps = int(self.cap.get(cv2.CAP_PROP_FPS))  # sudo apt-get install libv4l-dev
        # self.interval = int(self.fps / self.n_frame_per_sec)
        self.interval = 1. / self.n_frame_per_sec
        self.last_time = now()
        self.count = 0
        self.video_retry = 0
        logger.debug("Stream {} is started at fps={}".format(self.path, self.fps))

    def close(self):
        logger.debug("Closing video capture.")
        self.cap.release()

    def get_image(self):
        while (now() - self.last_time).total_seconds() < self.interval:
            time.sleep(0.01)

        self.count, frame = self.cap.read()

        self.last_time = now()

        if frame is None:
            if self.auto_restart is False:
                return None
            logger.error("Cannot get image from stream, tried {}".format(self.video_retry))
            self.video_retry += 1

            if self.video_retry > 3:
                logger.error(f"max_try {self.video_retry} has reached. Restarting")

                # restart stream
                self.close()
                self.start()

                return self.get_image()
            else:
                return self.get_image()

        return frame


@logger.catch(reraise=True)
def get_video_length(video_path):
    cap = cv2.VideoCapture(video_path)
    length = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    return length


@logger.catch(reraise=True)
def get_frames_from_video(video_path, start, end, interval):
    cap = cv2.VideoCapture(video_path)
    length = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    if start < 0:
        start = length + start
    elif start < 1:
        start = int(length * start)
    if end < 0:
        end = length + end
    elif end == 1:
        end = length
    elif end < 1:
        end = int(length * end)

    count = 1
    counts = []
    images = []
    while True:
        ret, frame = cap.read()
        if frame is not None and count <= end:
            if (interval == 1 or count % interval == 1) and start <= count <= end:
                counts.append(count)
                images.append(frame)
            count += 1
        else:
            break
    return counts, images


@logger.catch(reraise=True)
def get_frame_from_video(video_path, num_frame):
    cap = cv2.VideoCapture(video_path)
    length = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    if num_frame < 0:
        num_frame = length + num_frame
    elif num_frame < 1:
        num_frame = int(length * num_frame)
    count = 1
    while True:
        ret, image = cap.read()
        if image is not None:
            if count == num_frame:
                return image
            count += 1
        else:
            return None


@logger.catch(reraise=True)
def record_video_from_images(images, output_path):
    H, W = images[0].shape[:2]
    fps = 15
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    out = cv2.VideoWriter(output_path, fourcc, fps, (W, H))
    for image in images:
        out.write(image)

    out.release()


@logger.catch(reraise=True)
def record_video_from_queue(output_path, queue_image, fps=15):
    image = None
    # get the first roulette_image
    while image is None:
        image = queue_image.get(timeout=60)
    H, W = image.shape[:2]
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    out = cv2.VideoWriter(output_path, fourcc, fps, (W, H))

    while True:
        out.write(image)
        image = queue_image.get()
        if image is None:
            break

    out.release()


@logger.catch(reraise=True)
class Recorder:
    def __init__(self, output_path, fps=15):
        self.output_path = output_path
        self.queue_image = mp.Queue()
        self.p_record = mp.Process(target=record_video_from_queue, args=(self.output_path, self.queue_image, fps,))

    def start(self):
        self.p_record.start()
        logger.debug(f"Recorder output path {self.output_path}")

    def write(self, image):
        self.queue_image.put(image)

    def close(self):
        self.queue_image.put(None)
        self.queue_image.close()
        self.queue_image.join_thread()


if __name__ == '__main__':
    path = "../../sample/20191101-2_0758_0811_25.mp4"
    stream_gen = VideoYielder(path, 3, timeout=3)
    stream_gen.start()
    count = 0
    while True:
        image = stream_gen.get_image()
        if image is not None:
            count += 1
            if count % 30 == 1:
                print("get image {}".format(count))



# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: models.py 
@time: 2020/05/22
"""

# python packages
from absl import flags

# 3rd-party packages
import numpy as np
import tensorflow as tf
from tensorflow.keras.models import Model
from tensorflow.keras.layers import (
    Input,
    Lambda,
    MaxPool2D,
    Flatten
)

# self-defined packages
from cv.yolov3.core.layers import (
    DarknetConv,
    DarknetBlock,
    YoloConv,
    YoloConvTiny,
    YoloOutput,
    yolo_boxes,
    yolo_nms,
    stack_preds,
    multiclass_nms
)

flags.DEFINE_integer('yolo_max_boxes', 10, 'maximum number of boxes per image')
flags.DEFINE_float('yolo_iou_threshold', 0.45, 'iou threshold')
flags.DEFINE_float('yolo_score_threshold', 0.2, 'score threshold')

yolo_anchors = np.array([(10, 13), (16, 30), (33, 23), (30, 61), (62, 45),
                         (59, 119), (116, 90), (156, 198), (373, 326)],
                        np.float32) / 416
yolo_anchor_masks = np.array([[6, 7, 8], [3, 4, 5], [0, 1, 2]])

yolo_tiny_anchors = np.array([(10, 14), (23, 27), (37, 58),
                              (81, 82), (135, 169), (344, 319)],
                             np.float32) / 416
yolo_tiny_anchor_masks = np.array([[3, 4, 5], [0, 1, 2]])


def get_anchor_masks(anchors):
    if len(anchors) // 3 == 3:
        return yolo_anchor_masks
    else:
        return yolo_tiny_anchor_masks


def Darknet(name=None):
    x = inputs = Input([None, None, 3])
    x = Lambda(lambda _input: _input / 255.0)(x)
    x = DarknetConv(x, 32, 3)
    x = DarknetBlock(x, 64, 1)
    x = DarknetBlock(x, 128, 2)  # skip connection
    x = x_36 = DarknetBlock(x, 256, 8)  # skip connection
    x = x_61 = DarknetBlock(x, 512, 8)
    x = DarknetBlock(x, 1024, 4)
    return Model(inputs, (x_36, x_61, x), name=name)


def YoloV3(size=None, channels=3, anchors=yolo_anchors, masks=yolo_anchor_masks, classes=80, training=False):
    if isinstance(size, tuple):
        input_shape = size + (channels,)
    else:
        input_shape = (size, size, channels)

    x = inputs = Input(input_shape, name='input')

    x_36, x_61, x = Darknet(name='yolo_darknet')(x)

    x = YoloConv(512, name='yolo_conv_0')(x)
    output_0 = YoloOutput(512, len(masks[0]), classes, name='yolo_output_0')(x)

    x = YoloConv(256, name='yolo_conv_1')((x, x_61))
    output_1 = YoloOutput(256, len(masks[1]), classes, name='yolo_output_1')(x)

    x = YoloConv(128, name='yolo_conv_2')((x, x_36))
    output_2 = YoloOutput(128, len(masks[2]), classes, name='yolo_output_2')(x)

    if training:
        return Model(inputs, (output_0, output_1, output_2), name='yolov3')

    boxes_0 = Lambda(lambda x: yolo_boxes(x, anchors[masks[0]], classes), name='yolo_boxes_0')(output_0)
    boxes_1 = Lambda(lambda x: yolo_boxes(x, anchors[masks[1]], classes), name='yolo_boxes_1')(output_1)
    boxes_2 = Lambda(lambda x: yolo_boxes(x, anchors[masks[2]], classes), name='yolo_boxes_2')(output_2)

    outputs = Lambda(lambda x: yolo_nms(x, anchors, masks, classes), name='yolo_nms')(
        (boxes_0[:3], boxes_1[:3], boxes_2[:3]))

    return Model(inputs, outputs, name='yolov3')


def DarknetTiny(name=None):
    x = inputs = Input([None, None, 3])
    x = Lambda(lambda _input: _input / 255.0)(x)
    x = DarknetConv(x, 16, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = DarknetConv(x, 32, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = DarknetConv(x, 64, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = DarknetConv(x, 128, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = x_8 = DarknetConv(x, 256, 3)  # skip connection
    x = MaxPool2D(2, 2, 'same')(x)
    x = DarknetConv(x, 512, 3)
    x = MaxPool2D(2, 1, 'same')(x)
    x = DarknetConv(x, 1024, 3)
    return Model(inputs, (x_8, x), name=name)


def YoloV3Tiny(size=None, channels=3, anchors=yolo_tiny_anchors,
               masks=yolo_tiny_anchor_masks, classes=80, training=False):
    if isinstance(size, tuple):
        input_shape = size + (channels,)
    else:
        input_shape = (size, size, channels)

    x = inputs = Input(input_shape, name='input')

    x_8, x = DarknetTiny(name='yolo_darknet')(x)

    x = YoloConvTiny(256, name='yolo_conv_0')(x)
    output_0 = YoloOutput(256, len(masks[0]), classes, name='yolo_output_0')(x)

    x = YoloConvTiny(128, name='yolo_conv_1')((x, x_8))
    output_1 = YoloOutput(128, len(masks[1]), classes, name='yolo_output_1')(x)

    if training:
        return Model(inputs, (output_0, output_1), name='yolov3_tiny')

    boxes_0 = Lambda(lambda x: yolo_boxes(x, anchors[masks[0]], classes),
                     name='yolo_boxes_0')(output_0)
    boxes_1 = Lambda(lambda x: yolo_boxes(x, anchors[masks[1]], classes),
                     name='yolo_boxes_1')(output_1)
    outputs = Lambda(lambda x: yolo_nms(x, anchors, masks, classes),
                     name='yolo_nms')((boxes_0[:3], boxes_1[:3]))
    return Model(inputs, outputs, name='yolov3_tiny')


def YoloCustomizeModel(size=None, channels=3,
                       anchors=yolo_anchors, masks=None,
                       classes=80, training=False,
                       model=None,
                       normalize_input=True):
    """
    :param size:
    :param channels:
    :param anchors:
    :param masks:
    :param classes:
    :param training:
    :param model:
    :param normalize_input:
    :return:
    """
    # num_layer=2 --> YoloV3Tiny, num_layer=3 ---> YoloV3
    num_layer = len(anchors) // 3
    tiny = True if (num_layer == 2) else False
    if masks is None:
        masks = get_anchor_masks(anchors)

    if isinstance(size, tuple):
        input_shape = size + (channels,)
    else:
        input_shape = (size, size, channels)

    x = inputs = Input(input_shape, name='input')
    if normalize_input:
        x = Lambda(lambda _input: _input / 255.0)(x)

    if tiny:
        if model is None:
            x_skip, x = DarknetTiny(name='yolo_darknet')(x)
            x = y1 = YoloConvTiny(256, name='yolo_conv_0')(x)
            x = y2 = YoloConvTiny(128, name='yolo_conv_1')((x, x_skip))
        else:
            y1, y2 = model(x)

        output_0 = YoloOutput(256, len(masks[0]), classes, name='yolo_output_0')(y1)
        output_1 = YoloOutput(128, len(masks[1]), classes, name='yolo_output_1')(y2)

        outputs = [output_0, output_1]
    else:
        x_36, x_61, x = Darknet(name='yolo_darknet')(x)

        x = YoloConv(512, name='yolo_conv_0')(x)
        output_0 = YoloOutput(512, len(masks[0]), classes, name='yolo_output_0')(x)

        x = YoloConv(256, name='yolo_conv_1')((x, x_61))
        output_1 = YoloOutput(256, len(masks[1]), classes, name='yolo_output_1')(x)

        x = YoloConv(128, name='yolo_conv_2')((x, x_36))
        output_2 = YoloOutput(128, len(masks[2]), classes, name='yolo_output_2')(x)

        outputs = [output_0, output_1, output_2]

    if training:
        return Model(inputs, outputs, name='yolov3')

    boxes = [Lambda(lambda x: yolo_boxes(x, anchors[masks[i]], classes), name=f'yolo_boxes_{i}')(outputs[i]) for i in
             range(num_layer)]
    outputs = Lambda(lambda x: yolo_nms(x, anchors, masks, classes), name="yolo_nms")((box[:3] for box in boxes))

    return Model(inputs, outputs, name='yolov3_tiny')


def YoloCustomizeModelMultiCategory(size=None, channels=3,
                                    anchors=yolo_anchors, classes=[80], masks=None, training=False,
                                    model=None, normalize_input=True):
    """
    :param size:
    :param channels:
    :param anchors:
    :param masks:
    :param classes:
    :param training:
    :param model:
    :param normalize_input:
    :return:
    """
    # num_layer=2 --> YoloV3Tiny, num_layer=3 ---> YoloV3
    num_layer = len(anchors) // 3
    tiny = True if (num_layer == 2) else False
    if masks is None:
        masks = get_anchor_masks(anchors)

    if isinstance(size, tuple):
        input_shape = size + (channels,)
    else:
        input_shape = (size, size, channels)

    x = inputs = Input(input_shape, name='input')
    if normalize_input:
        x = Lambda(lambda _input: _input / 255.0)(x)

    if tiny:
        if model is None:
            x_skip, x = DarknetTiny(name='yolo_darknet')(x)
            x = y1 = YoloConvTiny(256, name='yolo_conv_0')(x)
            y2 = YoloConvTiny(128, name='yolo_conv_1')((x, x_skip))
        else:
            y1, y2 = model(x)

        output_0 = YoloOutput(256, len(masks[0]), classes, name='yolo_output_0')(y1)
        output_1 = YoloOutput(128, len(masks[1]), classes, name='yolo_output_1')(y2)
        outputs = [output_0, output_1]

    else:
        x_36, x_61, x = Darknet(name='yolo_darknet')(x)

        x = y1 = YoloConv(512, name='yolo_conv_0')(x)
        x = y2 = YoloConv(256, name='yolo_conv_1')((x, x_61))
        y3 = YoloConv(128, name='yolo_conv_2')((x, x_36))

        output_0 = YoloOutput(512, len(masks[0]), classes, name='yolo_output_0')(y1)
        output_1 = YoloOutput(256, len(masks[1]), classes, name='yolo_output_1')(y2)
        output_2 = YoloOutput(128, len(masks[2]), classes, name='yolo_output_2')(y3)
        outputs = [output_0, output_1, output_2]

    if training:
        return Model(inputs, outputs, name='yolov3')

    boxes = [Lambda(lambda x: yolo_boxes(x, anchors[masks[i]], classes),
                    name=f'yolo_boxes_{i}')(outputs[i]) for i in range(num_layer)]
    preds = Lambda(lambda x: stack_preds(x, classes), name="stack_pred")((box[:3] for box in boxes))
    outputs = Lambda(multiclass_nms)(preds)

    return Model(inputs, outputs, name='yolov3')



# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: label.py 
@time: 2020/02/24
"""

# python packages
import abc
import json

# 3rd-party packages
import cv2
import imgaug as ia
from loguru import logger

# self-defined packages
from utils import os_path, image_processing

ANNOTATION_FOLDER_DEFAULT_NAME = "Annotations"


class VOCJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'to_dic'):
            return obj.to_dic()

        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


class VOCLabel(metaclass=abc.ABCMeta):
    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.to_dic() == other.to_dic()
        return False

    @abc.abstractmethod
    def to_dic(self):
        return NotImplemented


class VOCBbox(VOCLabel):
    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def xmin(self):
        return self._xmin

    @xmin.setter
    def xmin(self, value):
        if hasattr(self, "xmax"):
            assert value < self.xmax, f"Cannot set property xmax, {value} > {self.xmax}"
        self._xmin = value
        if hasattr(self, "xmax"):
            self._width = self.xmax - self.xmin
            assert self.width > 0, f"width: {self._width} < 0"

    @property
    def x1(self):
        return self._xmin

    @property
    def x2(self):
        return self._xmax

    @property
    def y1(self):
        return self._ymin

    @property
    def y2(self):
        return self._ymax

    @property
    def xmax(self):
        return self._xmax

    @xmax.setter
    def xmax(self, value):
        if hasattr(self, "xmin"):
            assert value > self.xmin, f"Cannot set property xmin, {value} < {self.xmin}"
        self._xmax = value
        if hasattr(self, "xmin"):
            self._width = self.xmax - self.xmin
            assert self.width > 0 ,f"width: {self._width} < 0"

    @property
    def ymin(self):
        return self._ymin

    @ymin.setter
    def ymin(self, value):
        if hasattr(self, "ymax"):
            assert value < self.ymax, f"Cannot set property ymin, {value} > {self.ymax}"
        self._ymin = value
        if hasattr(self, "ymax"):
            self._height = self.ymax - self.ymin

    @property
    def ymax(self):
        return self._ymax

    @ymax.setter
    def ymax(self, value):
        self._ymax = value
        if hasattr(self, "ymin"):
            assert value > self.ymin, f"Cannot set property ymax, {value} < {self.ymin}"
            self._height = self.ymax - self.ymin

    def get_coords(self, format=None):
        if format is None:
            return self.x1, self.y1, self.x2, self.y2
        elif format == "dict":
            return {"x1": self.x1, "y1": self.y1, "x2": self.x2, "y2": self.y2}
        else:
            return (self.x1, self.y1), (self.x2, self.y2)

    def __init__(self, *args, **kargs):
        if args is None and kargs is None:
            xmin, ymin, xmax, ymax = 0, 0, 1, 1
        elif len(kargs) == 4:
            xmin, ymin, xmax, ymax = kargs["xmin"], kargs["ymin"], kargs["xmax"], kargs["ymax"]
        elif len(args) == 4:
            xmin, ymin, xmax, ymax = args
        else:
            raise Exception("Not enough args, kargs")

        self.xmin = int(xmin)
        self.xmax = int(xmax)
        self.ymin = int(ymin)
        self.ymax = int(ymax)

    def to_dic(self, recursively=True):
        if not recursively:
            return self
        return {"xmin": self.xmin, "ymin": self.ymin,
                "xmax": self.xmax, "ymax": self.ymax}

    def to_json_file(self, filepath):
        with open(filepath, "w+") as f:
            f.write(json.dumps(self.to_dic()))
        logger.debug(f"Json {filepath} exported.")

    def to_imgaug_bbox(self, label=None):
        imgaug_bbox = ia.BoundingBox(x1=self.xmin, y1=self.ymin, x2=self.xmax, y2=self.ymax, label=label)
        return imgaug_bbox

    @staticmethod
    def from_imgaug_bbox(imgaug_bbox):
        assert isinstance(imgaug_bbox, ia.BoundingBox)
        voc_bbox = VOCBbox(imgaug_bbox.x1, imgaug_bbox.y1, imgaug_bbox.x2, imgaug_bbox.y2)
        return voc_bbox


class VOCObject(VOCLabel):
    @property
    def truncated(self):
        return self._truncated

    @truncated.setter
    def truncated(self, value):
        assert isinstance(value, bool)
        self._truncated = value

    @property
    def difficult(self):
        return self._truncated

    @difficult.setter
    def difficult(self, value):
        assert isinstance(value, bool)
        self._difficult = value

    @property
    def defined_classes(self):
        return self._defined_classes

    @defined_classes.setter
    def defined_classes(self, value):
        self._defined_classes = value
        if value is None:
            return

        if isinstance(self.name, str):
            for defined_class in self.defined_classes:
                try:
                    self.name = defined_class.get_pair(self.name)
                    break
                except KeyError:
                    pass
        elif isinstance(self.name, list):
            for i, n in enumerate(self.name):
                try:
                    self.name[i] = self.defined_classes[i].get_pair(n)
                except KeyError:
                    pass

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def bbox(self):
        return self._bbox

    @bbox.setter
    def bbox(self, value):
        if isinstance(value, VOCBbox):
            self._bbox = value
        else:
            self._bbox = VOCBbox(*value)

    @property
    def x1(self):
        return self._bbox.x1

    @property
    def x2(self):
        return self._bbox.x2

    @property
    def y1(self):
        return self._bbox.y1

    @property
    def y2(self):
        return self._bbox.y2

    @property
    def score(self):
        return self._score

    @score.setter
    def score(self, value):
        self._score = value

    @property
    def from_annotation(self):  # parent
        return self._voc_annotation

    @from_annotation.setter
    def from_annotation(self, value):
        assert isinstance(value, VOCAnnotation) or value is None
        self._voc_annotation = value

    def __init__(self, name, bbox, truncated=False, difficult=False, defined_classes=None, from_annotation=None,
                 score=1.0):
        self.name = name
        self.bbox = bbox
        self.defined_classes = defined_classes
        self.truncated = truncated
        self.difficult = difficult
        self.from_annotation = from_annotation
        self.score = score

    def to_dic(self, recursively=True):
        if isinstance(self.name, list):
            name = [str(k) for k in self.name]
        else:
            name = str(self.name)
        return {"name": name,
                "bndbox": self.bbox.to_dic(recursively),
                "truncated": self.truncated,
                "difficult": self.difficult}

    def to_json_file(self, filepath):
        with open(filepath, "w+") as f:
            f.write(json.dumps(self.to_dic(recursively=True), cls=VOCJSONEncoder))
        logger.debug(f"Json {filepath} exported.")

    @staticmethod
    def load_json(j, defined_classes=None, from_annotation=None, debug=True):
        bbox = VOCBbox(**j["bndbox"])
        obj = VOCObject(name=j["name"], bbox=bbox, truncated=j["truncated"], difficult=j["difficult"],
                        defined_classes=defined_classes, from_annotation=from_annotation)
        if debug:
            logger.debug(f"VOCObject loaded from {j}")
        return obj

    def to_imgaug_bbox(self):
        imgaug_bbox = ia.BoundingBox(x1=self.bbox.xmin, y1=self.bbox.ymin, x2=self.bbox.xmax, y2=self.bbox.ymax,
                                     label=self.name)
        return imgaug_bbox

    @staticmethod
    def from_imgaug_bbox(imgaug_bbox):
        assert isinstance(imgaug_bbox, ia.BoundingBox)
        bbox = VOCBbox(imgaug_bbox.x1, imgaug_bbox.y1, imgaug_bbox.x2, imgaug_bbox.y2)
        voc_obj = VOCObject(imgaug_bbox.label, bbox)
        return voc_obj


class VOCAnnotation(VOCLabel):
    @property
    def image_path(self):  # image_path
        return self._image_path

    @image_path.setter
    def image_path(self, value):
        self._org_image_path = value
        self._image_path = value
        if self._image_path is not None:
            self._image_path = os_path.abspath(self._image_path)
            self._org_image_path = os_path.abspath(self._org_image_path)

    @property
    def annotation_path(self):
        if self._annotation_path:
            return self._annotation_path
        elif self.filename:  # default value
            return os_path.join(os_path.dirname(self.folder, full_path=True), ANNOTATION_FOLDER_DEFAULT_NAME,
                                self.filename + ".json")
        return None

    @annotation_path.setter
    def annotation_path(self, value):
        self._org_annotation_path = value
        self._annotation_path = value

    @property
    def filename(self):
        if self._annotation_path:
            return os_path.get_filename(self._annotation_path)
        if self.image_path:
            return os_path.get_filename(self.image_path)
        return None

    @property
    def folder(self):
        if self.image_path:
            return os_path.dirname(self.image_path)
            # return os_path.dirname(self.image_path, depth=2, full_path=False)
        if self.annotation_path:
            return os_path.dirname(self.annotation_path, depth=2)
            # return os_path.dirname(self.annotation_path, depth=2, full_path=False)
        return None

    @property
    def objects(self):
        return self._objects

    @property
    def bboxes(self):
        bboxes = [obj.bbox for obj in self.objects]
        return bboxes

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value):
        value_int = tuple([int(v) for v in value])
        self._size = value_int

    @property
    def segmented(self):
        return self._segmented

    @segmented.setter
    def segmented(self, value):
        self._segmented = value

    @property
    def defined_classes(self):
        return self._defined_classes

    @property
    def from_annotation_set(self):  # parent
        return self._from_annotation_set

    @from_annotation_set.setter
    def from_annotation_set(self, value):
        self._from_annotation_set = value

    @property
    def for_classification(self):  # parent
        return self._for_classification

    @for_classification.setter
    def for_classification(self, value):
        self._for_classification = value

    @property
    def unpacked(self):
        return self._unpacked

    def unpack(self):
        image = cv2.imread(self.image_path)
        self.size = image.shape
        self._unpacked = True

    def __init__(self, image_path=None, objects=None, annotation_path=None, from_annotation_set=None,
                 defined_classes=None, folder=None, filename=None, for_classification=False, size=None):
        # assert (image_path or (folder and filename))

        if not image_path and (folder and filename):
            image_path = os_path.join(folder, filename)

        self._objects = []
        if size:
            self._unpacked = True
        else:
            self._unpacked = False
        self.image_path = image_path
        self.size = size
        if objects:
            self.add_objects(objects)
        self.segmented = False
        self._defined_classes = defined_classes
        self.from_annotation_set = from_annotation_set
        self.annotation_path = annotation_path
        self.for_classification = for_classification

    def __len__(self):
        return len(self.objects)

    def add_objects(self, objs):
        if not isinstance(objs, list):
            objs = [objs]
        for obj in objs:
            assert isinstance(obj, VOCObject)
            self._objects.append(obj)

    def to_dic(self, recursively=True):
        if not self.size:
            size = (0, 0, 0)
        else:
            size = self.size
        if len(size) == 3:
            depth = size[2]
        else:
            depth = 1

        return {"folder": self.folder,
                "filename": os_path.basename(self.image_path),  # image filename
                "size": {"width": size[1],
                         "height": size[0],
                         "depth": depth},
                "objects": [obj.to_dic(recursively) for obj in self._objects],
                "segmented": self.segmented
                }

    def to_json_file(self, dir_=None, filepath=None):
        if dir_ is not None and filepath is None:
            filepath = os_path.join(dir_, self.filename)
        elif dir_ is None and filepath is None:
            filepath = self.annotation_path

        os_path.make_dir(os_path.dirname(filepath))

        with open(filepath, "w+") as f:
            f.write(json.dumps(self.to_dic(recursively=True), cls=VOCJSONEncoder))
        logger.debug(f"Json {filepath} exported.")

    @staticmethod
    def load_json(j, annotation_path=None, defined_classes=None, from_annotation_set=None, debug=True, image_dir=None):
        if image_dir is not None:
            image_path = os_path.join(image_dir, j["filename"])
        else:
            image_path = os_path.join(j["folder"], j["filename"])

        # check image path exists
        if not os_path.exists(image_path):
            if annotation_path:
                logger.warning(f"Image path {image_path} does not exist. Loading {annotation_path} failed.")
            else:
                logger.warning(f"Image path {image_path} does not exist. Loading {j} failed.")
            return None

        ann = VOCAnnotation(image_path=image_path, from_annotation_set=from_annotation_set,
                            annotation_path=annotation_path, defined_classes=defined_classes,
                            size=(j["size"]["height"], j["size"]["width"], j["size"]["depth"]))

        object_js = j["objects"]
        for object_j in object_js:
            obj = VOCObject.load_json(object_j, defined_classes=defined_classes, from_annotation=ann, debug=debug)
            ann.add_objects([obj])

        if "segmented" in j:
            ann.segmented = j["segmented"]

        if debug:
            logger.debug(f"VOC Annotation loaded from {j}")
        return ann

    @staticmethod
    def load_json_wrapper(j, queue, **kwargs):
        ann = VOCAnnotation.load_json(j, **kwargs)
        queue.put(ann)

    def to_imgaug_bbox_in_image(self):
        bbs = ia.BoundingBoxesOnImage([
            obj.to_imgaug_bbox() for obj in self.objects
        ], shape=self.size)
        return bbs

    def flatten(self, for_classification=True):
        annotations = []
        for obj in self.objects:
            ann = VOCAnnotation(image_path=self.image_path, annotation_path=self.annotation_path,
                                from_annotation_set=self.from_annotation_set, defined_classes=self.defined_classes,
                                objects=[obj], for_classification=for_classification, size=self.size)
            annotations.append(ann)
        return annotations

    def read_image(self):
        if os_path.exists(self.image_path):
            image = image_processing.imread(self.image_path)
        else:
            image = image_processing.imread(self._org_image_path)
        if self.for_classification:
            bbox = self.objects[0].bbox
            x1, y1, x2, y2 = bbox.x1, bbox.y1, bbox.x2, bbox.y2
            if len(image.shape) == 3:
                image = image[y1:y2 + 1, x1:x2 + 1, :]
            else:
                image = image[y1:y2 + 1, x1:x2 + 1]
        return image

    def load_from_json_file(self, json_file):
        with open(json_file) as f:
            j = json.load(f)
        self.__dict__.update(self.load_json(j, annotation_path=json_file).__dict__)


class VOCAnnotationSet(VOCLabel):
    @property
    def folder(self):
        self._folder = self.annotations[0].folder
        for annotation in self.annotations:
            if annotation.folder != self._folder:
                raise Exception("Not all annotation have same dir.")
        return self._folder

    @folder.setter
    def folder(self, value):
        logger.warning(f"Changing all annotations folder to {value}")
        for i, annotation in enumerate(self.annotations):
            self.annotations[i].folder = value

    @property
    def image_paths(self):  # image_paths
        return self._image_paths

    @property
    def annotations(self):
        return self._annotations

    @property
    def defined_classes(self):
        return self._defined_classes

    @property
    def class_count(self):
        return self._class_count

    @property
    def class_dict(self):
        return self._class_dict

    def __len__(self):
        return len(self.annotations)

    @property
    def unpacked(self):
        return self._unpacked

    @property
    def for_classification(self):
        return self._for_classification

    @for_classification.setter
    def for_classification(self, value):
        self._for_classification = value

    def __init__(self, dir_or_file=None, defined_classes=None, debug=False,
                 for_classification=False, image_dir=None):

        self._folder = ""
        self._image_paths = []
        self._annotations = []
        self._defined_classes = defined_classes
        self._class_count = {}
        self._class_dict = {}  # index of annotations
        self._unpacked = True
        self.for_classification = for_classification
        self.debug = debug

        # init class_count, class_dict
        if defined_classes:
            for defined_class in defined_classes:
                self._class_count[defined_class] = {}
                self._class_dict[defined_class] = {}
                for klass in list(defined_class):
                    self._class_count[defined_class][klass] = 0
                    self._class_dict[defined_class][klass] = []

        # load from dir or jsons or json
        if dir_or_file:
            if isinstance(dir_or_file, list) and all(os_path.isfile(x) for x in dir_or_file):
                self.load_from_jsons(sorted(dir_or_file), debug=debug,
                                     for_classification=for_classification, image_dir=image_dir)
            elif os_path.isdir(dir_or_file) or \
                    (isinstance(dir_or_file, list) and all(os_path.isdir(x) for x in dir_or_file)):
                if os_path.isdir(dir_or_file):
                    label_files = os_path.list_dir(dir_or_file, extension=".json")
                else:
                    label_files = []
                    for lf in dir_or_file:
                        label_files += os_path.list_dir(lf, extension=".json")
                if not label_files:
                    label_files = os_path.list_dir(os_path.join(dir_or_file, ANNOTATION_FOLDER_DEFAULT_NAME),
                                                   extension=".json")
                self.load_from_jsons(sorted(label_files), debug=debug, for_classification=for_classification,
                                     image_dir=image_dir)
            else:  # is file
                raise NotImplemented
                # self.load_from_json(dir_or_file)

    def add_annotation(self, annotation):
        if not annotation:
            return
        if isinstance(annotation, list):
            for ann in annotation:
                self.add_annotation(ann)
            return

        assert isinstance(annotation, VOCAnnotation)
        self._image_paths.append(annotation.image_path)
        self._annotations.append(annotation)
        self._unpacked = self._unpacked and annotation.unpacked

        # class_count
        if not self._defined_classes:
            for obj in annotation.objects:
                if isinstance(obj.name, list):
                    for n in obj.name:
                        if n in self._class_count:
                            self._class_count[n] += 1
                            self._class_dict[n].append(len(self))
                        else:
                            self._class_count[n] = 1
                            self._class_dict[n] = [len(self)]
                else:
                    if obj.name in self._class_count:
                        self._class_count[obj.name] += 1
                        self._class_dict[obj.name].append(len(self))
                    else:
                        self._class_count[obj.name] = 1
                        self._class_dict[obj.name] = [len(self)]
        else:
            for obj in annotation.objects:
                if isinstance(obj.name, list):
                    for n in obj.name:
                        self._class_count[type(n)][n] += 1
                        self._class_dict[type(n)][n].append(len(self))
                else:
                    self._class_count[type(obj.name)][obj.name] += 1
                    self._class_dict[type(obj.name)][obj.name].append(len(self))

    def to_dic(self, recursively=True):
        return [annotation.to_dic(recursively) for annotation in self.annotations]

    def to_json_files(self, dir_=None, filenames=None, filepaths=None):
        if filenames:
            assert len(filenames) == len(self._annotations)
        assert bool(dir_) is not bool(filepaths)
        if dir_:
            os_path.make_dir(dir_)
        for i, annotation in enumerate(self._annotations):
            if dir_ and filenames:
                annotation.to_json_file(filepath=os_path.join(dir_, filenames[i]))
            elif dir_:
                annotation.to_json_file(dir_=dir_)
            else:
                annotation.to_json_file()

    def to_json_file(self, filename):
        with open(filename, "w+") as f:
            f.write(json.dumps(self.to_dic(recursively=True), cls=VOCJSONEncoder))
        logger.debug(f"Json {filename} exported.")

    def load_from_jsons(self, label_files, debug=False, for_classification=False, image_dir=None):

        for label_file in label_files:
            with open(label_file, "r") as f:
                j = json.load(f)
            annotation = VOCAnnotation.load_json(j, annotation_path=label_file, from_annotation_set=self,
                                                 defined_classes=self.defined_classes, debug=debug,
                                                 image_dir=image_dir)
            if for_classification:
                if annotation:
                    annotation = annotation.flatten()
            self.add_annotation(annotation)

        logger.debug(f"Loaded {len(self)} annotations from jsons {label_files[-5:-1]}")
        return self

    def load_from_json(self, json_file):
        with open(json_file, "r") as f:
            annotation_js = json.load(f)

        for annotation_j in annotation_js:
            annotation = VOCAnnotation.load_json(annotation_j, self.defined_classes, from_annotation_set=self)
            self.add_annotation(annotation)
        logger.debug(f"Loaded {len(self)} annotations from {json_file}.")
        return self

    def unpack(self):
        for ann in self.annotations:
            ann.unpack()
        self._unpacked = True


